@@ -20,7 +20,6 @@
 -export([start/0, stop/0, action/5, diagnostics/1]).
 
 -define(RPC_TIMEOUT, infinity).
--define(WAIT_FOR_VM_ATTEMPTS, 5).
 
 -define(QUIET_OPT, "-q").
 -define(NODE_OPT, "-n").
@@ -191,9 +190,9 @@ action(force_cluster, Node, ClusterNodeSs, _Opts, Inform) ->
            [Node, ClusterNodes]),
     rpc_call(Node, rabbit_mnesia, force_cluster, [ClusterNodes]);
 
-action(wait, Node, [], _Opts, Inform) ->
+action(wait, Node, [PidFile], _Opts, Inform) ->
     Inform("Waiting for ~p", [Node]),
-    wait_for_application(Node, ?WAIT_FOR_VM_ATTEMPTS);
+    wait_for_application(Node, PidFile, Inform);
 
 action(status, Node, [], _Opts, Inform) ->
     Inform("Status of node ~p", [Node]),
@@ -235,17 +234,17 @@ action(clear_password, Node, Args = [Username], _Opts, Inform) ->
     Inform("Clearing password for user ~p", [Username]),
     call(Node, {rabbit_auth_backend_internal, clear_password, Args});
 
-action(set_admin, Node, [Username], _Opts, Inform) ->
-    Inform("Setting administrative status for user ~p", [Username]),
-    call(Node, {rabbit_auth_backend_internal, set_admin, [Username]});
-
-action(clear_admin, Node, [Username], _Opts, Inform) ->
-    Inform("Clearing administrative status for user ~p", [Username]),
-    call(Node, {rabbit_auth_backend_internal, clear_admin, [Username]});
+action(set_user_tags, Node, [Username | TagsStr], _Opts, Inform) ->
+    Tags = [list_to_atom(T) || T <- TagsStr],
+    Inform("Setting tags for user ~p to ~p", [Username, Tags]),
+    rpc_call(Node, rabbit_auth_backend_internal, set_tags,
+             [list_to_binary(Username), Tags]);
 
 action(list_users, Node, [], _Opts, Inform) ->
     Inform("Listing users", []),
-    display_list(call(Node, {rabbit_auth_backend_internal, list_users, []}));
+    display_info_list(
+      call(Node, {rabbit_auth_backend_internal, list_users, []}),
+      rabbit_auth_backend_internal:user_info_keys());
 
 action(add_vhost, Node, Args = [_VHostPath], _Opts, Inform) ->
     Inform("Creating vhost ~p", Args),
@@ -301,7 +300,7 @@ action(list_connections, Node, Args, _Opts, Inform) ->
 
 action(list_channels, Node, Args, _Opts, Inform) ->
     Inform("Listing channels", []),
-    ArgAtoms = default_if_empty(Args, [pid, user, transactional, consumer_count,
+    ArgAtoms = default_if_empty(Args, [pid, user, consumer_count,
                                        messages_unacknowledged]),
     display_info_list(rpc_call(Node, rabbit_channel, info_all, [ArgAtoms]),
                       ArgAtoms);
@@ -354,23 +353,63 @@ action(report, Node, _Args, _Opts, Inform) ->
 
 %%----------------------------------------------------------------------------
 
-wait_for_application(Node, Attempts) ->
-    case rpc_call(Node, application, which_applications, [infinity]) of
-        {badrpc, _} = E -> case Attempts of
-                               0 -> E;
-                               _ -> wait_for_application0(Node, Attempts - 1)
-                           end;
-        Apps            -> case proplists:is_defined(rabbit, Apps) of
-                               %% We've seen the node up; if it goes down
-                               %% die immediately.
-                               true  -> ok;
-                               false -> wait_for_application0(Node, 0)
-                           end
+wait_for_application(Node, PidFile, Inform) ->
+    Pid = wait_and_read_pid_file(PidFile),
+    Inform("pid is ~s", [Pid]),
+    wait_for_application(Node, Pid).
+
+wait_for_application(Node, Pid) ->
+    case process_up(Pid) of
+        true  -> case rabbit:is_running(Node) of
+                     true  -> ok;
+                     false -> timer:sleep(1000),
+                              wait_for_application(Node, Pid)
+                 end;
+        false -> {error, process_not_running}
     end.
 
-wait_for_application0(Node, Attempts) ->
-    timer:sleep(1000),
-    wait_for_application(Node, Attempts).
+wait_and_read_pid_file(PidFile) ->
+    case file:read_file(PidFile) of
+        {ok,    Bin}    -> string:strip(binary_to_list(Bin), right, $\n);
+        {error, enoent} -> timer:sleep(500),
+                           wait_and_read_pid_file(PidFile);
+        {error, _} = E  -> exit({error, {could_not_read_pid, E}})
+    end.
+
+% Test using some OS clunkiness since we shouldn't trust
+% rpc:call(os, getpid, []) at this point
+process_up(Pid) ->
+    with_os([{unix, fun () ->
+                            system("ps -p " ++ Pid
+                                   ++ " >/dev/null 2>&1") =:= 0
+                    end},
+             {win32, fun () ->
+                             Res = os:cmd("tasklist /nh /fi \"pid eq " ++
+                                          Pid ++ "\" 2>&1"),
+                             case re:run(Res, "erl\\.exe", [{capture, none}]) of
+                                 match -> true;
+                                 _     -> false
+                             end
+                     end}]).
+
+with_os(Handlers) ->
+    {OsFamily, _} = os:type(),
+    case proplists:get_value(OsFamily, Handlers) of
+        undefined -> throw({unsupported_os, OsFamily});
+        Handler   -> Handler()
+    end.
+
+% Like system(3)
+system(Cmd) ->
+    ShCmd = "sh -c '" ++ escape_quotes(Cmd) ++ "'",
+    Port = erlang:open_port({spawn, ShCmd}, [exit_status,nouse_stdio]),
+    receive {Port, {exit_status, Status}} -> Status end.
+
+% Escape the quotes in a shell command so that it can be used in "sh -c 'cmd'"
+escape_quotes(Cmd) ->
+    lists:flatten(lists:map(fun ($') -> "'\\''"; (Ch) -> Ch end, Cmd)).
+
+%%----------------------------------------------------------------------------
 
 default_if_empty(List, Default) when is_list(List) ->
     if List == [] -> Default;
@@ -422,17 +461,6 @@ format_info_item([T | _] = Value)
 format_info_item(Value) ->
     io_lib:format("~w", [Value]).
 
-display_list(L) when is_list(L) ->
-    lists:foreach(fun (I) when is_binary(I) ->
-                          io:format("~s~n", [escape(I)]);
-                      (I) when is_tuple(I) ->
-                          display_row([escape(V)
-                                       || V <- tuple_to_list(I)])
-                  end,
-                  lists:sort(L)),
-    ok;
-display_list(Other) -> Other.
-
 display_call_result(Node, MFA) ->
     case call(Node, MFA) of
         {badrpc, _} = Res -> throw(Res);
